package xyz.lychee.lagfixer.managers;

import com.mojang.authlib.GameProfile;
import com.mojang.authlib.properties.Property;
import com.sun.management.OperatingSystemMXBean;
import java.lang.management.ManagementFactory;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;
import org.apache.commons.lang.RandomStringUtils;
import org.bukkit.Bukkit;
import org.bukkit.Chunk;
import org.bukkit.GameRule;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.entity.Arrow;
import org.bukkit.entity.Creature;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Item;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Projectile;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.HandlerList;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.CreatureSpawnEvent;
import org.bukkit.event.entity.EntityDeathEvent;
import org.bukkit.event.entity.ItemDespawnEvent;
import org.bukkit.event.entity.ItemSpawnEvent;
import org.bukkit.event.entity.ProjectileHitEvent;
import org.bukkit.event.entity.ProjectileLaunchEvent;
import org.bukkit.event.world.ChunkLoadEvent;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.plugin.Plugin;
import xyz.lychee.lagfixer.LagFixer;
import xyz.lychee.lagfixer.modules.CustomAiModule;
import xyz.lychee.lagfixer.modules.CustomVehiclesModule;
import xyz.lychee.lagfixer.objects.AbstractFork;
import xyz.lychee.lagfixer.objects.AbstractManager;
import xyz.lychee.lagfixer.objects.AbstractMonitor;
import xyz.lychee.lagfixer.objects.AbstractStacker;
import xyz.lychee.lagfixer.objects.AbstractSupportNms;
import xyz.lychee.lagfixer.utils.ReflectionUtils;

public class SupportManager extends AbstractManager implements Listener {
   public static final Map<String, EntityType> ENTITIES = (Map)Arrays.stream(EntityType.values()).collect(Collectors.toMap(Enum::name, Function.identity()));
   public static final Map<String, String> VERSIONS = new HashMap();
   public static final ScheduledExecutorService EXECUTOR;
   private static SupportManager instance;
   private final Map<World, HashSet<LivingEntity>> entityMap = new ConcurrentHashMap();
   private final Map<World, HashSet<Item>> itemMap = new ConcurrentHashMap();
   private final Map<World, HashSet<Projectile>> projectileMap = new ConcurrentHashMap();
   private ScheduledFuture<?> task = null;
   private AbstractFork fork = null;
   private AbstractSupportNms nms = null;
   private AbstractStacker stacker = null;
   private AbstractMonitor monitor = new StandardMonitor();

   public SupportManager(LagFixer plugin) {
      super(plugin);
      instance = this;
   }

   public HashSet<LivingEntity> getCreatureSet(World world) {
      return (HashSet)this.entityMap.computeIfAbsent(world, (k) -> {
         return new HashSet();
      });
   }

   public HashSet<Item> getItemSet(World world) {
      return (HashSet)this.itemMap.computeIfAbsent(world, (k) -> {
         return new HashSet();
      });
   }

   public HashSet<Projectile> getProjectileSet(World world) {
      return (HashSet)this.projectileMap.computeIfAbsent(world, (k) -> {
         return new HashSet();
      });
   }

   public int getCreatureCount() {
      return this.entityMap.values().stream().mapToInt(Set::size).sum();
   }

   public int getItemCount() {
      return this.itemMap.values().stream().mapToInt(Set::size).sum();
   }

   public int getProjectileCount() {
      return this.projectileMap.values().stream().mapToInt(Set::size).sum();
   }

   public void load() {
      Bukkit.getPluginManager().registerEvents(this, this.getPlugin());
      Bukkit.getWorlds().forEach((world) -> {
         this.entityMap.put(world, new HashSet(world.getLivingEntities()));
         this.itemMap.put(world, new HashSet(world.getEntitiesByClass(Item.class)));
      });
      this.getPlugin().getLogger().info(" &8• &rLoaded creatures " + this.getCreatureCount() + " to creature container!");
      this.getPlugin().getLogger().info(" &8• &rLoaded items " + this.getItemCount() + " to item container!");

      try {
         this.fork = (AbstractFork)Arrays.stream(new Class[]{Class.forName("xyz.lychee.lagfixer.support.Spigot"), Class.forName("xyz.lychee.lagfixer.support.Folia")}).map((clazzx) -> {
            try {
               return (AbstractFork)clazzx.getConstructor(Plugin.class).newInstance(this.getPlugin());
            } catch (Exception var3) {
               return null;
            }
         }).filter(Objects::nonNull).filter(AbstractFork::isSupported).max(Comparator.comparingInt(AbstractFork::getPriority)).orElse((Object)null);
         if (this.fork != null) {
            this.getPlugin().getLogger().info(" &8• &rLoaded fork support ~ " + this.fork.getClass().getCanonicalName());
         }

         String version = ReflectionUtils.getVersion("SupportNms");
         Class<?> clazz = Class.forName("xyz.lychee.lagfixer.nms." + version + ".SupportNms");
         this.nms = (AbstractSupportNms)clazz.getConstructor(Plugin.class).newInstance(this.getPlugin());
      } catch (Exception var3) {
         this.getPlugin().getLogger().info("   &cOptimal support not found, the plugin will use deprecated methods!");
         this.getPlugin().getLogger().info("   &7Supported versions: &e1.16.5, 1.17.1, 1.18.2, 1.19.4, 1.20.x, 1.21-1.21.3");
         this.nms = new DeprecatedBukkitSupport(this.getPlugin());
      }

      Bukkit.getPluginManager().registerEvents(this.nms, this.getPlugin());
      this.getPlugin().getLogger().info(" &8• &rLoaded nms support ~ " + this.nms.getClass().getCanonicalName());
      this.task = EXECUTOR.scheduleAtFixedRate(() -> {
         this.entityMap.values().forEach((set) -> {
            set.removeIf((ent) -> {
               return !ent.isValid();
            });
         });
         this.itemMap.values().forEach((set) -> {
            set.removeIf((ent) -> {
               return !ent.isValid();
            });
         });
      }, 5L, 5L, TimeUnit.SECONDS);
   }

   public void disable() {
      HandlerList.unregisterAll(this);
      if (this.task != null) {
         this.task.cancel(true);
      }

   }

   public boolean isEnabled() {
      return true;
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onCreatureSpawn(CreatureSpawnEvent event) {
      EXECUTOR.execute(() -> {
         HashSet<LivingEntity> set = this.getCreatureSet(event.getLocation().getWorld());
         set.add(event.getEntity());
      });
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onItemSpawn(ItemSpawnEvent event) {
      EXECUTOR.execute(() -> {
         HashSet<Item> set = this.getItemSet(event.getLocation().getWorld());
         set.add(event.getEntity());
      });
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onLaunch(ProjectileLaunchEvent event) {
      EXECUTOR.execute(() -> {
         HashSet<Projectile> set = this.getProjectileSet(event.getLocation().getWorld());
         set.add(event.getEntity());
      });
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onDespawn(ItemDespawnEvent event) {
      EXECUTOR.execute(() -> {
         HashSet<Item> set = this.getItemSet(event.getLocation().getWorld());
         set.remove(event.getEntity());
      });
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onDeath(EntityDeathEvent event) {
      EXECUTOR.execute(() -> {
         HashSet<LivingEntity> set = this.getCreatureSet(event.getEntity().getWorld());
         set.remove(event.getEntity());
      });
   }

   @EventHandler(
      priority = EventPriority.HIGHEST,
      ignoreCancelled = true
   )
   public void onHit(ProjectileHitEvent event) {
      EXECUTOR.execute(() -> {
         if (!(event.getEntity() instanceof Arrow)) {
            HashSet<Projectile> set = this.getProjectileSet(event.getEntity().getWorld());
            set.remove(event.getEntity());
         }
      });
   }

   public Map<World, HashSet<LivingEntity>> getEntityMap() {
      return this.entityMap;
   }

   public Map<World, HashSet<Item>> getItemMap() {
      return this.itemMap;
   }

   public Map<World, HashSet<Projectile>> getProjectileMap() {
      return this.projectileMap;
   }

   public ScheduledFuture<?> getTask() {
      return this.task;
   }

   public AbstractFork getFork() {
      return this.fork;
   }

   public AbstractSupportNms getNms() {
      return this.nms;
   }

   public AbstractStacker getStacker() {
      return this.stacker;
   }

   public AbstractMonitor getMonitor() {
      return this.monitor;
   }

   public void setTask(ScheduledFuture<?> task) {
      this.task = task;
   }

   public void setFork(AbstractFork fork) {
      this.fork = fork;
   }

   public void setNms(AbstractSupportNms nms) {
      this.nms = nms;
   }

   public void setStacker(AbstractStacker stacker) {
      this.stacker = stacker;
   }

   public void setMonitor(AbstractMonitor monitor) {
      this.monitor = monitor;
   }

   public static SupportManager getInstance() {
      return instance;
   }

   static {
      VERSIONS.put("1.20.5", "v1_20_R4");
      VERSIONS.put("1.20.6", "v1_20_R4");
      VERSIONS.put("1.21", "v1_21_R1");
      VERSIONS.put("1.21.1", "v1_21_R1");
      VERSIONS.put("1.21.2", "v1_21_R2");
      VERSIONS.put("1.21.3", "v1_21_R2");
      EXECUTOR = Executors.newSingleThreadScheduledExecutor();
   }

   class StandardMonitor extends AbstractMonitor {
      private final OperatingSystemMXBean osBean = (OperatingSystemMXBean)ManagementFactory.getPlatformMXBean(OperatingSystemMXBean.class);

      protected double cpuProcess() {
         return this.osBean.getProcessCpuLoad() * 100.0;
      }

      protected double cpuSystem() {
         return this.osBean.getSystemCpuLoad() * 100.0;
      }

      protected double tps() {
         return SupportManager.this.getNms().getTps(0);
      }

      protected double tps(int time) {
         return SupportManager.this.getNms().getTps(time);
      }

      protected double mspt() {
         return SupportManager.this.getFork().getMspt();
      }
   }

   class DeprecatedBukkitSupport extends AbstractSupportNms {
      public DeprecatedBukkitSupport(Plugin plugin) {
         super(plugin);
      }

      public double getTps(int time) {
         return 20.0;
      }

      public void setDistance(int view, int simulation) {
      }

      public void setDefaultDistance() {
      }

      public int getViewDistance() {
         return Bukkit.getViewDistance();
      }

      public int getViewSimulation() {
         return -1;
      }

      public ItemStack setNBTValue(ItemStack itemStack, String key, String value) {
         return itemStack;
      }

      public boolean hasNBTValue(ItemStack itemStack, String key) {
         return false;
      }

      public String getNBTValue(ItemStack itemStack, String key) {
         return null;
      }

      public ItemStack createSkull(String base64) {
         ItemStack is = new ItemStack(Material.PLAYER_HEAD);
         ItemMeta meta = is.getItemMeta();
         if (meta == null) {
            return is;
         } else {
            GameProfile profile = new GameProfile(UUID.randomUUID(), RandomStringUtils.randomAlphabetic(8));
            profile.getProperties().put("textures", new Property("textures", base64));

            try {
               Method mtd = meta.getClass().getDeclaredMethod("setProfile", GameProfile.class);
               mtd.setAccessible(true);
               mtd.invoke(meta, profile);
            } catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException var6) {
               return is;
            }

            is.setItemMeta(meta);
            return is;
         }
      }

      public void setEntityAi(Entity ent, boolean enable) {
         if (ent instanceof Creature) {
            Creature c = (Creature)ent;
            if (c.hasAI() == enable) {
               return;
            }

            c.setAI(enable);
            c.setSilent(!enable);
            c.setCollidable(enable);
         }

      }

      public void setRandomTickSpeed(World world, boolean enable) {
         SupportManager.this.getFork().runSync((Location)null, () -> {
            world.setGameRule(GameRule.RANDOM_TICK_SPEED, enable ? 3 : 0);
         });
      }

      public int getTileEntitiesCount(Chunk chunk) {
         return 0;
      }

      public int averagePing() {
         return -1;
      }

      public void setSpawnLimits(World world, int monsters, int animals, int waterAnimals, int waterAmbient, int waterUndergroundCreature, int axolotls, int ambient) {
         world.setMonsterSpawnLimit(monsters);
         world.setAnimalSpawnLimit(animals);
         world.setWaterAnimalSpawnLimit(waterAnimals);
         world.setAmbientSpawnLimit(ambient);
      }

      @EventHandler
      public void onLoad(ChunkLoadEvent e) {
         SupportManager.EXECUTOR.execute(() -> {
            Entity[] entities = e.getChunk().getEntities();
            CustomAiModule module1 = (CustomAiModule)ModuleManager.getInstance().get(CustomAiModule.class);
            int var7;
            if (module1 != null && module1.canContinue(e.getWorld()) && module1.isLoaded()) {
               CustomAiModule.NMS nmsx = module1.getCustomAi();
               Entity[] var5 = entities;
               int var6 = entities.length;

               for(var7 = 0; var7 < var6; ++var7) {
                  Entity entx = var5[var7];
                  if (module1.isEnabled(entx)) {
                     nmsx.setCustomAi(entx, false);
                  }
               }
            }

            CustomVehiclesModule module2 = (CustomVehiclesModule)ModuleManager.getInstance().get(CustomVehiclesModule.class);
            if (module2 != null && module2.canContinue(e.getWorld()) && module2.isLoaded()) {
               CustomVehiclesModule.NMS nms = module2.getCustomVehicles();
               Entity[] var15 = entities;
               var7 = entities.length;

               for(int var18 = 0; var18 < var7; ++var18) {
                  Entity entxx = var15[var18];
                  if (module2.isEnabled(entxx)) {
                     nms.setCustomVehicle(entxx);
                  }
               }
            }

            Set<LivingEntity> creatures = SupportManager.this.getCreatureSet(e.getWorld());
            Set<Item> items = SupportManager.this.getItemSet(e.getWorld());
            Set<Projectile> projectiles = SupportManager.this.getProjectileSet(e.getWorld());
            Entity[] var19 = entities;
            int var20 = entities.length;

            for(int var10 = 0; var10 < var20; ++var10) {
               Entity ent = var19[var10];
               if (ent instanceof LivingEntity) {
                  creatures.add((LivingEntity)ent);
               } else if (ent instanceof Item) {
                  items.add((Item)ent);
               } else if (ent instanceof Projectile) {
                  projectiles.add((Projectile)ent);
               }
            }

         });
      }
   }
}
